import wx
from wx.lib.pubsub import Publisher
from math import log10,exp,pow

from peak_o_mat import module
from peak_o_mat.plotcanvas import Canvas, Graphics, Line, Marker
from peak_o_mat import settings as config

import numpy as N

class Module(module.Module):
    title = 'Ruby Calibration'
    
    def __init__(self, *args):
        module.Module.__init__(self, __file__, *args)
        
    def init(self):
        self.Bind(wx.EVT_TEXT, self.OnTemp, self.xrc_txt_temp)
        Publisher().subscribe(self.new_model, ('fitfinished'))
        self.plotcanvas = Canvas(self.panel, style=wx.SUNKEN_BORDER)
        self.plotcanvas.SetShowScrollbars(False)
        self.plotcanvas.antialias = config.options.antialias
        self.xmlres.AttachUnknownControl('xrc_pan_plot', self.plotcanvas)
        self.panel.Layout()

        self.plot()
        
    def plot(self, point=None):

        lines = []

        x = N.linspace(4,300,100)
        y = 1.6067e-3*N.power(x,1.565)
        lines.append(Line(N.transpose([x,y])))
        if point is not None:
            lines.append(Marker([point], size=2, marker='circle', colour='red'))
        
        gr = Graphics(lines, 'melting curve of He4','temperature (K)', 'pressure (GPa)')
        self.plotcanvas.Draw(gr, (0,300),(0,12))
        
    def calculate(self, model):
        ne_line = 6929.47

        r1 = None
        ne = None

        for f in model:
            if f.name.find('GA') == 0:
                if ne is None:
                    ne = f.pos.value
                else:
                    #self.message('more than one ne line found', blink=True)
                    return False
        if ne is None:
            return False
        
        for f in model:
            if f.name.find('LO') == 0:
                r1 = max(r1,f.pos.value)
        if r1 is None:
            self.message('ruby r1 line not found', blink=True)
            return False

        if log10(ne) < 3.0:
            ne *= 10
            r1 *= 10

        self.xrc_txt_r1.SetValue('%.2f'%r1)
        self.xrc_txt_neon.SetValue('%.2f'%ne)
        
        try:
            temp = float(self.xrc_txt_temp.GetValue())
        except ValueError:
            return False
        
        r1 = r1+(ne_line-ne)
        try:
            p = self.pressure(temp, r1)
        except:
            return False
        else:
            self.xrc_txt_press.SetValue('%.2f'%p)
            wx.CallAfter(self.plot, (temp, p))
            return True
        
    def pressure(self, t, r1):
        """
        calculates the hydrostatic pressure from temperature and
        the wavelength of the r1 line in Angstroem
        parameters are taken from Ph.D. thesis of Clemens Ulrich
        """
        A = 1904
        B = 7.665
        alpha = 76.6
        theta =  482
        r1_0 = 14421.8

        e = 1/(r1*1e-8)

        r1_t = r1_0 - alpha/(exp(theta/t)-1)

        r1_t = 1/(1e-8*r1_t)

        p = A/B*(pow(1+(r1*1e-10-r1_t*1e-10)/(r1_t*1e-10),B)-1)

        return p

    def update(self):
        aset = self.controller.active_set
        if aset is not None and aset.mod is not None:
            self.panel.Enable()
            if not self.calculate(aset.mod):
                self.xrc_txt_press.SetValue('???')
            else:
                self.message('')
        else:
            self.panel.Disable()
            self.plot()

    def OnTemp(self, evt):
        self.update()
        
    def selection_changed(self):
        self.update()
            
    def new_model(self, msg):
        self.selection_changed()

    def page_changed(self, me):
        if me:
            self.selection_changed()
